cimport cython
cimport numpy
import numpy as np
import matplotlib.pyplot as plt
from scipy.io import wavfile
from scipy.signal import detrend, lfilter, bilinear, spectrogram, filtfilt, resample, fftconvolve
import acoustics
import itertools

from acoustics.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
from acoustics.standards.iec_61672_1_2013 import WEIGHTING_SYSTEMS
from acoustics.standards.iec_61672_1_2013 import (NOMINAL_OCTAVE_CENTER_FREQUENCIES,
                                                  NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES)


class Signal(numpy.ndarray):
    """A signal consisting of samples (array) and a sample frequency (float).

    """
    #fs = 0.0

    def __new__(cls, data, fs):

        #if np.asarray(data).ndim!=1:
            #raise ValueError("Incorrect amount of dimensions. One dimension is required.")

        obj = np.asarray(data).view(cls)
        #obj = np.atleast_2d(np.asarray(data)).view(cls)
        obj.fs = fs #if fs is not None else 1000
        return obj

    def __array_prepare__(self, array, context=None):
        try:
            a = context[1][0]
            b = context[1][1]
        except IndexError:
            return array

        if hasattr(a, 'fs') and hasattr(b, 'fs'):
            if a.fs == b.fs:
                return array
            else:
                raise ValueError("Sample frequencies do not match.")
        else:
            return array

    def __array_wrap__(self, out_arr, context=None):
        return np.ndarray.__array_wrap__(self, out_arr, context)

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        #attr = getattr(obj, 'fs', None)
        #print(attr)
        #self.fs = attr if attr is not None else 44100.0
        self.fs = getattr(obj, 'fs', None)#44100.0)
        #self.fs = 1000

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Signal, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.fs,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.fs = state[-1]  # Set the info attribute
        # Call the parent's __setstate__ with the other tuple elements.
        super(Signal, self).__setstate__(state[0:-1])

    def __repr__(self):
        return "Signal({})".format(str(self))

    def _construct(self, x):
        """Construct signal like x."""
        return Signal(x, self.fs)

    @property
    def samples(self):
        """Amount of samples in signal."""
        return self.shape[-1]

    @property
    def channels(self):
        """Amount of channels.
        """
        if self.ndim > 1:
            return self.shape[-2]
        else:
            return 1

    @property
    def duration(self):
        """Duration of signal in seconds.
        """
        return self.samples / self.fs

    @property
    def values(self):
        """Return the values of this signal as an instance of :class:`np.ndarray`."""
        return np.array(self)

    def calibrate_to(self, decibel, inplace=False):
        """Calibrate signal to value `decibel`.

        :param decibel: Value to calibrate to.
        :param inplace: Whether to perform inplace or not.
        :returns: Calibrated signal.
        :rtype: :class:`Signal`

        Values of `decibel` are broadcasted. To set a value per channel, use `decibel[...,None]`.
        """
        decibel = decibel * np.ones(self.shape)
        gain = decibel - self.leq()[...,None]
        return self.gain(gain, inplace=inplace)

    def calibrate_with(self, other, decibel, inplace=False):
        """Calibrate signal with other signal.

        :param other: Other signal/array.
        :param decibel: Signal level of `other`.
        :param inplace: Whether to perform inplace or not.
        :returns: calibrated signal.
        :rtype: :class:`Signal`
        """
        if not isinstance(other, Signal):
            other = Signal(other, self.fs)
        gain = decibel - other.leq()
        return self.gain(gain, inplace=inplace)

    def decimate(self, factor, zero_phase=False, ftype='iir', order=None):
        """Decimate signal by integer `factor`. Before downsampling a low-pass filter is applied.

        :param factor: Downsampling factor.
        :param zero_phase: Prevent phase shift by filtering with ``filtfilt`` instead of ``lfilter``.
        :param ftype: Filter type.
        :param order: Filter order.
        :returns: Decimated signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`scipy.signal.decimate`
        .. seealso:: :meth:`resample`

        """
        return Signal(acoustics.signal.decimate(x=self, q=factor, n=order, ftype=ftype, zero_phase=zero_phase), self.fs/factor)

    def resample(self, nsamples, times=None, axis=-1, window=None):
        """Resample signal.

        :param samples:: New amount of samples.
        :param times: Times corresponding to samples.
        :param axis: Axis.
        :param window: Window.

        .. seealso:: :func:`scipy.signal.resample`
        .. seealso:: :meth:`decimate`

        You might want to low-pass filter this signal before resampling.

        """
        return Signal(resample(self, nsamples, times, axis, window), nsamples/self.samples*self.fs)

    def upsample(self, factor, axis=-1):
        """Upsample signal with integer factor.

        :param factor: Upsample factor.
        :param axis: Axis.

        .. seealso:: :meth:`resample`
        """
        return self.resample(int(self.samples*factor), axis=-1)


    def gain(self, decibel, inplace=False):
        """Apply gain of `decibel` decibels.

        :param decibel: Decibels
        :param inplace: In place
        :returns: Amplified signal.
        :rtype: :class:`Signal`
        """
        factor = 10.0**(decibel/20.0)
        if inplace:
            self *= factor
            return self
        else:
            return self * factor

    def pick(self, start=0.0, stop=None):
        """Get signal from start time to stop time.

        :param start: Start time.
        :type start: float
        :param stop: End time.
        :type stop: float
        :returns: Selected part of the signal.
        :rtype: :class:`Signal`

        """
        if start is not None:
            start = np.floor(start*self.fs)
        if stop is not None:
            stop  = np.floor(stop*self.fs)
        return self[..., start:stop]

    def times(self):
        """Time vector.

        :returns: A vector with a timestamp for each sample.
        :rtype: :class:`np.ndarray`

        """
        return np.arange(0, self.samples) / self.fs

    def energy(self):
        """Signal energy.

        :returns: Total energy per channel.
        :rtype: :class:`np.ndarray`

        .. math:: E = \\sum_{n=0}^{N-1} |x_n|^2

        """
        return float((self*self).sum())

    def power(self):
        """Signal power.

        .. math:: P = \\frac{1}{N} \\sum_{n=0}^{N-1} |x_n|^2
        """
        return self.energy() / len(self)

    def ms(self):
        """Mean value squared of signal.

        .. seealso:: `acoustics.signal.ms`

        """
        return acoustics.signal.ms(self)

    def rms(self):
        """Root mean squared of signal.

        .. seealso:: `acoustics.signal.rms`

        """
        return acoustics.signal.rms(self)
        #return np.sqrt(self.power())


    def weigh(self, weighting='A', zero_phase=False):
        """Apply frequency-weighting. By default 'A'-weighting is applied.

        :param weighting: Frequency-weighting filter to apply. Valid options are 'A', 'C' and 'Z'.
        :returns: A-weighted signal.
        :rtype: :class:`Signal`.

        By default the weighting filter is applied using :func:`scipy.signal.lfilter` causing a frequency-dependent delay.
        In case a delay is undesired, the filter can applied using :func:`scipy.signal.filtfilt` by settings `zero_phase=True`.

        """
        num, den = WEIGHTING_SYSTEMS[weighting]()
        b, a = bilinear(num, den, self.fs)
        if zero_phase:
            self._construct(filtfilt(b, a, self))
        else:
            return self._construct(lfilter(b, a, self))


    def correlate(self, other=None, mode='full'):
        """Correlate signal with `other` signal. In case `other==None` this
        method returns the autocorrelation.

        :param other: Other signal.
        :param mode: Mode.

        .. seealso:: :func:`np.correlate`, :func:`scipy.signal.fftconvolve`

        """
        if other is None:
            other = self
        if self.fs != other.fs:
            raise ValueError("Cannot correlate. Sample frequencies are not the same.")
        if self.channels > 1 or other.channels > 1:
            raise ValueError("Cannot correlate. Not supported for multichannel signals.")
        return self._construct(fftconvolve(self, other[::-1], mode=mode))

    def amplitude_envelope(self):
        """Amplitude envelope.

        :returns: Amplitude envelope of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`acoustics.signal.amplitude_envelope`

        """
        return self._construct(acoustics.signal.amplitude_envelope(self, self.fs))

    def instantaneous_frequency(self):
        """Instantaneous frequency.

        :returns: Instantaneous frequency of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`acoustics.signal.instantaneous_frequency`

        """
        return self._construct(acoustics.signal.instantaneous_frequency(self, self.fs))

    def instantaneous_phase(self):
        """Instantaneous phase.

        :returns: Instantaneous phase of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`acoustics.signal.instantaneous_phase`

        """
        return self._construct(acoustics.signal.instantaneous_phase(self, self.fs))


    def detrend(self, **kwargs):
        """Detrend signal.

        :returns: Detrended version of signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`scipy.signal.detrend`

        """
        return self._construct(detrend(self, **kwargs))

    def unwrap(self):
        """Unwrap signal in case the signal represents wrapped phase.

        :returns: Unwrapped signal.
        :rtype: :class:`Signal`

        .. seealso:: :func:`np.unwrap`

        """
        return self._construct(np.unwrap(self))

    def complex_cepstrum(self, N=None):
        """Complex cepstrum.

        :param N: Amount of bins.
        :returns: Quefrency, complex cepstrum and delay in amount of samples.

        .. seealso:: :func:`acoustics.cepstrum.complex_cepstrum`

        """
        if N is not None:
            times = np.linspace(0.0, self.duration, N, endpoint=False)
        else:
            times = self.times()
        cepstrum, ndelay = acoustics.cepstrum.complex_cepstrum(self, n=N)
        return times, cepstrum, ndelay


    def real_cepstrum(self, N=None):
        """Real cepstrum.

        :param N: Amount of bins.
        :returns: Quefrency and real cepstrum.

        .. seealso:: :func:`acoustics.cepstrum.real_cepstrum`

        """
        if N is not None:
            times = np.linspace(0.0, self.duration, N, endpoint=False)
        else:
            times = self.times()
        return times, acoustics.cepstrum.real_cepstrum(self, n=N)


    def power_spectrum(self, N=None):
        """Power spectrum.

        :param N: Amount of bins.

        .. seealso:: :func:`acoustics.signal.power_spectrum`

        """
        return acoustics.signal.power_spectrum(self, self.fs, N=N)


    def angle_spectrum(self, N=None):
        """Phase angle spectrum. Wrapped.

        :param N: amount of bins.

        .. seealso:: :func:`acoustics.signal.angle_spectrum`, :func:`acoustics.signal.phase_spectrum` and :meth:`phase_spectrum`.

        """
        return acoustics.signal.angle_spectrum(self, self.fs, N=N)

    def phase_spectrum(self, N=None):
        """Phase spectrum. Unwrapped.

        :param N: Amount of bins.

        .. seealso:: :func:`acoustics.signal.phase_spectrum`, :func:`acoustics.signal.angle_spectrum` and :meth:`angle_spectrum`.

        """
        return acoustics.signal.phase_spectrum(self, self.fs, N=N)


    def peak(self, axis=-1):
        """Peak sound pressure.

        :param axis: Axis.

        .. seealso:: :func:`acoustic.standards.iso_tr_25417_2007.peak_sound_pressure`

        """
        return acoustics.standards.iso_tr_25417_2007.peak_sound_pressure(self, axis=axis)


    def peak_level(self, axis=-1):
        """Peak sound pressure level.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.peak_sound_pressure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.peak_sound_pressure_level(self, axis=axis)


    def min(self, axis=-1):
        """Return the minimum along a given axis.

        Refer to `numpy.amin` for full documentation.
        """
        return numpy.ndarray.min(self, axis=axis)


    def max(self, axis=-1):
        """Return the minimum along a given axis.

        Refer to `numpy.amax` for full documentation.
        """
        return numpy.ndarray.max(self, axis=axis)


    def max_level(self, axis=-1):
        """Maximum sound pressure level.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.max_sound_pressure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.max_sound_pressure_level(self, axis=axis)


    def sound_exposure(self, axis=-1):
        """Sound exposure.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.sound_exposure`

        """
        return acoustics.standards.iso_tr_25417_2007.sound_exposure(self, self.fs, axis=axis)


    def sound_exposure_level(self, axis=-1):
        """Sound exposure level.

        :param axis: Axis.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.sound_exposure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.sound_exposure_level(self, self.fs, axis=axis)


    def plot_complex_cepstrum(self, N=None, **kwargs):
        """Plot complex cepstrum of signal.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * frequency: Boolean indicating whether the x-axis should show time in seconds or quefrency
        * xlabel_frequency: Label in case frequency is shown.

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$t$ in s",
            'ylabel': "$C$",
            'title' : 'Complex cepstrum',
            'frequency' : False,
            'xlabel_frequency' : "$f$ in Hz",
            }
        params.update(kwargs)

        t, ceps, _ = self.complex_cepstrum(N=N)
        if params['frequency']:
            t = 1./t
            params['xlabel'] = params['xlabel_frequency']
            t = t[::-1]
            ceps = ceps[::-1]
        return _base_plot(t, ceps, params)


    def plot_real_cepstrum(self, N=None, **kwargs):
        """Plot real cepstrum of signal.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * frequency: Boolean indicating whether the x-axis should show time in seconds or quefrency
        * xlabel_frequency: Label in case frequency is shown.

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$t$ in s",
            'ylabel': "$C$",
            'title' : 'Real cepstrum',
            'frequency' : False,
            'xlabel_frequency' : "$f$ in Hz",
            }
        params.update(kwargs)

        t, ceps = self.real_cepstrum(N=N)
        if params['frequency']:
            t = 1./t
            params['xlabel'] = params['xlabel_frequency']
            t = t[::-1]
            ceps = ceps[::-1]
        return _base_plot(t, ceps, params)


    def plot_power_spectrum(self, N=None, **kwargs):#filename=None, scale='log'):
        """Plot spectrum of signal.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * reference: Reference power

        .. seealso:: :meth:`power_spectrum`

        """
        params = {
            'xscale': 'log',
            'yscale': 'linear',
            'xlabel': "$f$ in Hz",
            'ylabel': "$L_{p}$ in dB",
            'title' : 'SPL',
            'reference' : REFERENCE_PRESSURE**2.0,
            }
        params.update(kwargs)

        f, o = self.power_spectrum(N=N)
        return _base_plot(f, 10.0*np.log10(o/params['reference']), params)


    def plot_angle_spectrum(self, N=None, **kwargs):
        """Plot phase angle spectrum of signal. Wrapped.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * reference: Reference power

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$f$ in Hz",
            'ylabel': "$\\angle \phi$",
            'title' : 'Phase response (wrapped)',
            }
        params.update(kwargs)
        f, o = self.angle_spectrum(N=N)
        return _base_plot(f, o, params)


    def plot_phase_spectrum(self, N=None, **kwargs):
        """Plot phase spectrum of signal. Unwrapped.

        Valid kwargs:

        * xscale
        * yscale
        * xlim
        * ylim
        * reference: Reference power

        """
        params = {
            'xscale': 'linear',
            'yscale': 'linear',
            'xlabel': "$f$ in Hz",
            'ylabel': "$\\angle \phi$",
            'title' : 'Phase response (unwrapped)',
            }
        params.update(kwargs)
        f, o = self.phase_spectrum(N=N)
        return _base_plot(f, o, params)


    def spectrogram(self, **kwargs):
        """Spectrogram of signal.

        :returns: Spectrogram.

        See :func:`scipy.signal.spectrogram`. Some of the default values have been changed.
        The generated spectrogram consists by default of complex values.

        """
        params = {
            'nfft' : 4096,
            'noverlap' : 128,
            'mode' : 'complex',
            }
        params.update(kwargs)

        t, s, P = spectrogram(self, fs=self.fs, **params)

        return t, s, P


    def plot_spectrogram(self, **kwargs):
        """
        Plot spectrogram of the signal.

        Valid kwargs:

        * xlim
        * ylim
        * clim
        .. note:: This method only works for a single channel.

        """
        # To do, use :meth:`spectrogram`.
        params = {
            'xlim' : None,
            'ylim' : None,
            'clim' : None,
            'NFFT' : 4096,
            'noverlap' : 128,
            'title': 'Spectrogram',
            'xlabel': '$t$ in s',
            'ylabel': '$f$ in Hz',
            'clabel': 'SPL in dB',
            }
        params.update(kwargs)

        if self.channels > 1:
            raise ValueError("Cannot plot spectrogram of multichannel signal. Please select a single channel.")
        fig = plt.figure()
        ax0 = fig.add_subplot(111)
        ax0.set_title(params['title'])
        #f = ax0.specgram(self, Fs=self.fs)
        data = np.squeeze(self)
        try:
            _, _, _, im = ax0.specgram(data, Fs=self.fs, noverlap=params['noverlap'], NFFT=params['NFFT'], mode='magnitude', scale_by_freq=False)
        except AttributeError:
            raise NotImplementedError("Your version of matplotlib is incompatible due to lack of support of the mode keyword argument to matplotlib.mlab.specgram.")

        cb = fig.colorbar(mappable=im)
        cb.set_label(params['clabel'])

        ax0.set_xlim(params['xlim'])
        ax0.set_ylim(params['ylim'])
        im.set_clim(params['clim'])

        ax0.set_xlabel(params['xlabel'])
        ax0.set_ylabel(params['ylabel'])

        return fig


    def levels(self, time=0.125, method='average'):
        """Calculate sound pressure level as function of time.

        :param time: Averaging time or integration time constant. Default value is 0.125 corresponding to FAST.
        :param method: Use time `average` or time `weighting`. Default option is `average`.
        :returns: sound pressure level as function of time.

        .. seealso:: :func:`acoustics.standards.iec_61672_1_2013.time_averaged_sound_level`
        .. seealso:: :func:`acoustics.standards.iec_61672_1_2013.time_weighted_sound_level`

        """
        if method=='average':
            return acoustics.standards.iec_61672_1_2013.time_averaged_sound_level(self.values, self.fs, time)
        elif method=='weighting':
            return acoustics.standards.iec_61672_1_2013.time_weighted_sound_level(self.values, self.fs, time)
        else:
            raise ValueError("Invalid method")

    def leq(self):
        """Equivalent level. Single-value number.

        .. seealso:: :func:`acoustics.standards.iso_tr_25417_2007.equivalent_sound_pressure_level`

        """
        return acoustics.standards.iso_tr_25417_2007.equivalent_sound_pressure_level(self.values)

    def plot_levels(self, **kwargs):
        """Plot sound pressure level as function of time.

        .. seealso:: :meth:`levels`

        """
        params = {
            'xscale'    :   'linear',
            'yscale'    :   'linear',
            'xlabel'    :   '$t$ in s',
            'ylabel'    :   '$L_{p,F}$ in dB',
            'title'     :   'SPL',
            'time'      :   0.125,
            'method'    :   'average',
            'labels'    :   None,
            }
        params.update(kwargs)
        t, L = self.levels(params['time'], params['method'])
        L_masked = np.ma.masked_where(np.isinf(L), L)
        return _base_plot(t, L_masked, params)

    #def octave(self, frequency, fraction=1):
        #"""Determine fractional-octave `fraction` at `frequency`.

        #.. seealso:: :func:`acoustics.signal.fractional_octaves`

        #"""
        #return acoustics.signal.fractional_octaves(self, self.fs, frequency,
                                                  #frequency, fraction, False)[1]

    def bandpass(self, lowcut, highcut, order=8, zero_phase=False):
        """Filter signal with band-pass filter.

        :param lowcut: Lower cornerfrequency.
        :param highcut: Upper cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        :returns: Band-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.bandpass`
        """
        return type(self)(acoustics.signal.bandpass(self, lowcut, highcut, self.fs, order=order, zero_phase=zero_phase), self.fs)

    def bandstop(self, lowcut, highcut, order=8, zero_phase=False):
        """Filter signal with band-stop filter.

        :param lowcut: Lower cornerfrequency.
        :param highcut: Upper cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).

        :returns: Band-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.bandstop`
        """
        return type(self)(acoustics.signal.bandstop(self, lowcut, highcut, self.fs, order=order, zero_phase=zero_phase), self.fs)

    def highpass(self, cutoff, order=4, zero_phase=False):
        """Filter signal with high-pass filter.

        :param cutoff: Cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: High-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.highpass`
        """
        return type(self)(acoustics.signal.highpass(self, cutoff, self.fs, order=order, zero_phase=zero_phase), self.fs)

    def lowpass(self, cutoff, order=4, zero_phase=False):
        """Filter signal with low-pass filter.

        :param cutoff: Cornerfrequency.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Low-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.lowpass`
        """
        return type(self)(acoustics.signal.lowpass(self, cutoff, self.fs, order=order, zero_phase=zero_phase), self.fs)

    def octavepass(self, center, fraction, order=8, zero_phase=False):
        """Filter signal with fractional-octave band-pass filter.

        :param center: Center frequency. Any value in the band will suffice.
        :param fraction: Band designator.
        :param order: Filter order.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Band-pass filtered signal.
        :rtype: :class:`Signal`.

        .. seealso:: :func:`acoustics.signal.octavepass`
        """
        return type(self)(acoustics.signal.octavepass(self, center, self.fs, fraction=fraction, order=order, zero_phase=zero_phase), self.fs)

    def bandpass_frequencies(self, frequencies, order=8, purge=True, zero_phase=False):
        """Apply bandpass filters for frequencies.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: Instance of :class:`acoustics.signal.Frequencies`
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_frequencies`
        """
        frequencies, filtered = acoustics.signal.bandpass_frequencies(self, self.fs, frequencies, order, purge, zero_phase=zero_phase)
        return frequencies, type(self)(filtered, self.fs)


    def octaves(self, frequencies=NOMINAL_OCTAVE_CENTER_FREQUENCIES, order=8, purge=True, zero_phase=False):
        """Apply 1/1-octaves bandpass filters.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: :class:`np.ndarray` with (approximate) center-frequencies or an instance of :class:`acoustics.signal.Frequencies`
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_octaves`
        """
        frequencies, octaves = acoustics.signal.bandpass_octaves(self, self.fs, frequencies, order, purge, zero_phase=zero_phase)
        return frequencies, type(self)(octaves, self.fs)


    def third_octaves(self, frequencies=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES, order=8, purge=True, zero_phase=False):
        """Apply 1/3-octaves bandpass filters.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: :class:`np.ndarray` with (approximate) center-frequencies or an instance of :class:`acoustics.signal.Frequencies`
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_third_octaves`
        """
        frequencies, octaves = acoustics.signal.bandpass_third_octaves(self, self.fs, frequencies, order, purge, zero_phase=zero_phase)
        return frequencies, type(self)(octaves, self.fs)


    def fractional_octaves(self, frequencies=None, fraction=1, order=8, purge=True, zero_phase=False):
        """Apply 1/N-octaves bandpass filters.

        :param frequencies: Band-pass filter frequencies.
        :type frequencies: Instance of :class:`acoustics.signal.Frequencies`
        :param fraction: Default band-designator of fractional-octaves.
        :param order: Filter order.
        :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
        :param zero_phase: Prevent phase error by filtering in both directions (filtfilt).
        :returns: Frequencies and band-pass filtered signal.

        .. seealso:: :func:`acoustics.signal.bandpass_fractional_octaves`
        """
        if frequencies is None:
            frequencies = acoustics.signal.OctaveBand(fstart=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES[0], fstop=self.fs/2.0, fraction=fraction)
        frequencies, octaves = acoustics.signal.bandpass_fractional_octaves(self, self.fs, frequencies, fraction, order, purge, zero_phase=zero_phase)
        return frequencies, type(self)(octaves, self.fs)


    def plot_octaves(self, **kwargs):
        """Plot octaves.

        .. seealso:: :meth:`octaves`

        """
        params = {
            'xscale'    :   'log',
            'yscale'    :   'linear',
            'xlabel'    :   '$f$ in Hz',
            'ylabel'    :   '$L_{p}$ in dB',
            'title'     :   '1/1-Octaves SPL',
            }
        params.update(kwargs)
        f, o = self.octaves()
        print(len(f.center), len(o.leq()))
        return _base_plot(f.center, o.leq().T, params)

    def plot_third_octaves(self, **kwargs):
        """Plot 1/3-octaves.

        .. seealso:: :meth:`third_octaves`

        """
        params = {
            'xscale'    :   'log',
            'yscale'    :   'linear',
            'xlabel'    :   '$f$ in Hz',
            'ylabel'    :   '$L_{p}$ in dB',
            'title'     :   '1/3-Octaves SPL',
            }
        params.update(kwargs)
        f, o = self.third_octaves()
        return _base_plot(f.center, o.leq().T, params)

    def plot_fractional_octaves(self, frequencies=None, fraction=1, order=8, purge=True, zero_phase=False, **kwargs):
        """Plot fractional octaves.
        """
        title = '1/{}-Octaves SPL'.format(fraction)

        params = {
            'xscale'    :   'log',
            'yscale'    :   'linear',
            'xlabel'    :   '$f$ in Hz',
            'ylabel'    :   '$L_p$ in dB',
            'title'     :   title,
        }
        params.update(kwargs)
        f, o = self.fractional_octaves(frequencies=frequencies, fraction=fraction, order=order, purge=purge, zero_phase=zero_phase)
        return _base_plot(f.center, o.leq().T, params)

    def plot(self, **kwargs):
        """Plot signal as function of time. By default the entire signal is plotted.

        :param filename: Name of file.
        :param start: First sample index.
        :type start: Start time in seconds from start of signal.
        :param stop: Last sample index.
        :type stop: Stop time in seconds. from stop of signal.
        """
        params = {
            'xscale'    :   'linear',
            'yscale'    :   'linear',
            'xlabel'    :   '$t$ in s',
            'ylabel'    :   '$x$ in -',
            'title'     :   'Signal',
            }
        params.update(kwargs)
        return _base_plot(self.times(), self, params)

    #def plot_scalo(self, filename=None):
        #"""
        #Plot scalogram
        #"""
        #from scipy.signal import ricker, cwt

        #wavelet = ricker
        #widths = np.logspace(-1, 3.5, 10)
        #x = cwt(self, wavelet, widths)

        #interpolation = 'nearest'

        #from matplotlib.ticker import LinearLocator, AutoLocator, MaxNLocator
        #majorLocator = LinearLocator()
        #majorLocator = MaxNLocator()


        #fig = plt.figure()
        #ax = fig.add_subplot(111)
        #ax.set_title('Scaleogram')
        ##ax.set_xticks(np.arange(0, x.shape[1])*self.fs)
        ##ax.xaxis.set_major_locator(majorLocator)

        ##ax.imshow(10.0 * np.log10(x**2.0), interpolation=interpolation, aspect='auto', origin='lower')#, extent=[0, 1, 0, len(x)])
        #ax.pcolormesh(np.arange(0.0, x.shape[1])/self.fs, widths, 10.0*np.log(x**2.0))
        #if filename:
            #fig.savefig(filename)
        #else:
            #return fig

    #def plot_scaleogram(self, filename):
        #"""
        #Plot scaleogram
        #"""
        #import pywt

        #wavelet = 'dmey'
        #level = pywt.dwt_max_level(len(self), pywt.Wavelet(wavelet))
        #print level
        #level = 20
        #order = 'freq'
        #interpolation = 'nearest'

        #wp = pywt.WaveletPacket(self, wavelet, 'sym', maxlevel=level)
        #nodes = wp.get_level(level, order=order)
        #labels = [n.path for n in nodes]
        #values = np.abs(np.array([n.data for n in nodes], 'd'))

        #fig = plt.figure()
        #ax = fig.add_subplot(111)
        #ax.set_title('Scaleogram')
        #ax.imshow(values, interpolation=interpolation, aspect='auto', origin='lower', extent=[0, 1, 0, len(values)])
        ##ax.set_yticks(np.arange(0.5, len(labels) + 0.5))
        ##ax.set_yticklabels(labels)

        #fig.savefig(filename)

    def normalize(self, gap=6.0, inplace=False):
        """Normalize signal.

        :param gap: Gap between maximum value and ceiling in decibel.
        :param inplace: Normalize signal in place.

        The parameter `gap` can be understood as using `gap` decibels fewer for the dynamic range.
        By default a 6 decibel gap is used.

        """
        factor = (self.max() * 10.0**(gap/20.0))
        if inplace:
            self /= factor
            return self
        else:
            return self / factor


    def to_wav(self, filename, depth=16):
        """Save signal as WAV file.

        :param filename: Name of file to save to.
        :param depth: If given, convert to integer with specified depth. Else, try to store using the original data type.

        By default, this function saves a normalized 16-bit version of the signal with at least 6 dB range till clipping occurs.

        """
        data = self
        dtype = data.dtype if not depth else 'int'+str(depth)
        if depth:
            data = (data * 2**(depth-1)-1).astype(dtype)
        wavfile.write(filename, int(self.fs), data.T)
        #wavfile.write(filename, int(self.fs), self._data/np.abs(self._data).max() *  0.5)
        #wavfile.write(filename, int(self.fs), np.int16(self._data/(np.abs(self._data).max()) * 32767) )

    @classmethod
    def from_wav(cls, filename):
        """
        Create an instance of `Signal` from a WAV file.

        :param filename: Filename

        """
        fs, data = wavfile.read(filename)
        data = data.astype(np.float32, copy=False).T
        data /= np.max(np.abs(data))
        return cls(data, fs=fs)


_PLOTTING_PARAMS = {
    'title'     :   None,
    'xlabel'    :   None,
    'ylabel'    :   None,
    'xscale'    :   'linear',
    'yscale'    :   'linear',
    'xlim'      :   (None, None),
    'ylim'      :   (None, None),
    'labels'    :   None,
    'linestyles':   ['-', '-.', '--', ':'],
    }

def _get_plotting_params():
    d = dict()
    d.update(_PLOTTING_PARAMS)
    return d


def _base_plot(x, y, given_params):

    params = _get_plotting_params()
    params.update(given_params)

    linestyles = itertools.cycle(iter(params['linestyles']))

    fig = plt.figure()
    ax0 = fig.add_subplot(111)
    ax0.set_title(params['title'])
    if y.ndim > 1:
        for channel in y:
            ax0.plot(x, channel, linestyle=next(linestyles))
    else:
        ax0.plot(x, y)
    ax0.set_xlabel(params['xlabel'])
    ax0.set_ylabel(params['ylabel'])
    ax0.set_xscale(params['xscale'])
    ax0.set_yscale(params['yscale'])
    ax0.set_xlim(params['xlim'])
    ax0.set_ylim(params['ylim'])

    if params['labels'] is None and y.ndim > 1:
        params['labels'] = np.arange(y.shape[-2])+1
    if params['labels'] is not None:
        ax0.legend(labels=params['labels'])

    return fig



"""Cepstrum.
"""

import numpy as np

__all__ = ['complex_cepstrum', 'real_cepstrum', 'inverse_complex_cepstrum',
           'minimum_phase']
 
def complex_cepstrum(x, n=None):
    """Compute the complex cepstrum of a real sequence.
        
    Parameters
    ----------
    x : ndarray
        Real sequence to compute complex cepstrum of.
    n : {None, int}, optional
        Length of the Fourier transform.
    
    Returns
    -------
    ceps : ndarray
        The complex cepstrum of the real data sequence `x` computed using the 
        Fourier transform.
    ndelay : int
        The amount of samples of circular delay added to `x`.
    
    The complex cepstrum is given by
    
    .. math:: c[n] = F^{-1}\\left{\\log_{10}{\\left(F{x[n]}\\right)}\\right}
    
    where :math:`x_[n]` is the input signal and :math:`F` and :math:`F_{-1} 
    are respectively the forward and backward Fourier transform.

    See Also
    --------
    real_cepstrum: Compute the real cepstrum.
    inverse_complex_cepstrum: Compute the inverse complex cepstrum of a real sequence.

    
    Examples
    --------
    In the following example we use the cepstrum to determine the fundamental 
    frequency of a set of harmonics. There is a distinct peak at the quefrency 
    corresponding to the fundamental frequency. To be more precise, the peak 
    corresponds to the spacing between the harmonics.
    
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from scipy.signal import complex_cepstrum
    
    >>> duration = 5.0
    >>> fs = 8000.0
    >>> samples = int(fs*duration)
    >>> t = np.arange(samples) / fs
    
    >>> fundamental = 100.0
    >>> harmonics = np.arange(1, 30) * fundamental
    >>> signal = np.sin(2.0*np.pi*harmonics[:,None]*t).sum(axis=0)
    >>> ceps, _ = complex_cepstrum(signal)
    
    >>> fig = plt.figure()
    >>> ax0 = fig.add_subplot(211)
    >>> ax0.plot(t, signal)
    >>> ax0.set_xlabel('time in seconds')
    >>> ax0.set_xlim(0.0, 0.05)
    >>> ax1 = fig.add_subplot(212)
    >>> ax1.plot(t, ceps)
    >>> ax1.set_xlabel('quefrency in seconds')
    >>> ax1.set_xlim(0.005, 0.015)
    >>> ax1.set_ylim(-5., +10.)
    
    References
    ----------
    .. [1] Wikipedia, "Cepstrum".
           http://en.wikipedia.org/wiki/Cepstrum
    .. [2] M.P. Norton and D.G. Karczub, D.G., 
           "Fundamentals of Noise and Vibration Analysis for Engineers", 2003.
    .. [3] B. P. Bogert, M. J. R. Healy, and J. W. Tukey: 
           "The Quefrency Analysis of Time Series for Echoes: Cepstrum, Pseudo
           Autocovariance, Cross-Cepstrum and Saphe Cracking". 
           Proceedings of the Symposium on Time Series Analysis
           Chapter 15, 209-243. New York: Wiley, 1963.
    
    """
    def _unwrap(phase):
        samples = phase.shape[-1]
        unwrapped = np.unwrap(phase)
        center = (samples+1)//2
        if samples == 1: 
            center = 0  
        ndelay = np.array(np.round(unwrapped[...,center]/np.pi))
        unwrapped -= np.pi * ndelay[...,None] * np.arange(samples) / center
        return unwrapped, ndelay
        
    spectrum = np.fft.fft(x, n=n)
    unwrapped_phase, ndelay = _unwrap(np.angle(spectrum))
    log_spectrum = np.log(np.abs(spectrum)) + 1j*unwrapped_phase
    ceps = np.fft.ifft(log_spectrum).real
    
    return ceps, ndelay


def real_cepstrum(x, n=None):
    """Compute the real cepstrum of a real sequence. 
    
    x : ndarray
        Real sequence to compute real cepstrum of.
    n : {None, int}, optional
        Length of the Fourier transform.
    
    Returns
    -------
    ceps: ndarray
        The real cepstrum.
    
    The real cepstrum is given by
    
    .. math:: c[n] = F^{-1}\\left{\\log_{10}{\\left|F{x[n]}\\right|}\\right}
    
    where :math:`x_[n]` is the input signal and :math:`F` and :math:`F_{-1} 
    are respectively the forward and backward Fourier transform. Note that 
    contrary to the complex cepstrum the magnitude is taken of the spectrum.
    
    
    See Also
    --------
    complex_cepstrum: Compute the complex cepstrum of a real sequence.
    inverse_complex_cepstrum: Compute the inverse complex cepstrum of a real sequence.
    
    Examples
    --------
    >>> from scipy.signal import real_cepstrum
    
    
    References
    ----------
    .. [1] Wikipedia, "Cepstrum".
           http://en.wikipedia.org/wiki/Cepstrum
    
    """
    spectrum = np.fft.fft(x, n=n)
    ceps = np.fft.ifft(np.log(np.abs(spectrum))).real
    
    return ceps


def inverse_complex_cepstrum(ceps, ndelay):
    """Compute the inverse complex cepstrum of a real sequence.
    
    ceps : ndarray
        Real sequence to compute inverse complex cepstrum of.
    ndelay: int
        The amount of samples of circular delay added to `x`.
    
    Returns
    -------
    x : ndarray
        The inverse complex cepstrum of the real sequence `ceps`.
    
    The inverse complex cepstrum is given by
    
    .. math:: x[n] = F^{-1}\\left{\\exp(F(c[n]))\\right}
    
    where :math:`c_[n]` is the input signal and :math:`F` and :math:`F_{-1} 
    are respectively the forward and backward Fourier transform.
    
    See Also
    --------
    complex_cepstrum: Compute the complex cepstrum of a real sequence.
    real_cepstrum: Compute the real cepstrum of a real sequence.
      
    Examples
    --------
    Taking the complex cepstrum and then the inverse complex cepstrum results 
    in the original sequence.
    
    >>> import numpy as np
    >>> from scipy.signal import inverse_complex_cepstrum
    >>> x = np.arange(10)
    >>> ceps, ndelay = complex_cepstrum(x)
    >>> y = inverse_complex_cepstrum(ceps, ndelay)
    >>> print(x)
    >>> print(y)

    References
    ----------
    .. [1] Wikipedia, "Cepstrum".
           http://en.wikipedia.org/wiki/Cepstrum
      
    """
    def _wrap(phase, ndelay):
        ndelay = np.array(ndelay)
        samples = phase.shape[-1]
        center = (samples+1)//2
        wrapped = phase + np.pi * ndelay[...,None] * np.arange(samples) / center
        return wrapped
    
    log_spectrum = np.fft.fft(ceps)
    spectrum = np.exp(log_spectrum.real + 1j * _wrap(log_spectrum.imag, ndelay))
    x = np.fft.ifft(spectrum).real
    return x


def minimum_phase(x, n=None):
    """Compute the minimum phase reconstruction of a real sequence.
    
    x : ndarray
        Real sequence to compute the minimum phase reconstruction of.    
    n : {None, int}, optional
        Length of the Fourier transform.
    
    Compute the minimum phase reconstruction of a real sequence using the 
    real cepstrum.
    
    Returns
    -------
    m : ndarray
        The minimum phase reconstruction of the real sequence `x`.
    
    See Also
    --------
    real_cepstrum: Compute the real cepstrum.
    
    Examples
    --------
    >>> from scipy.signal import minimum_phase
    
    
    References
    ----------
    .. [1] Soo-Chang Pei, Huei-Shan Lin. Minimum-Phase FIR Filter Design Using 
           Real Cepstrum. IEEE TRANSACTIONS ON CIRCUITS AND SYSTEMS-II: 
           EXPRESS BRIEFS, VOL. 53, NO. 10, OCTOBER 2006
    
    """
    if n is None:
        n = len(x)
    ceps = real_cepstrum(x, n=n)
    odd = n % 2 
    window = np.concatenate(([1.0], 2.0*np.ones((n+odd)/2-1), 
                             np.ones(1-odd), np.zeros((n+odd)/2-1)))
    
    m = np.fft.ifft(np.exp(np.fft.fft(window*ceps))).real
    
    return m



"""
Acoustics
=========

The acoustics module.

"""
import acoustics.aio
import acoustics.ambisonics
import acoustics.atmosphere
import acoustics.bands
import acoustics.building
import acoustics.cepstrum
import acoustics.criterion
import acoustics.decibel
import acoustics.descriptors
import acoustics.directivity
import acoustics.doppler
import acoustics.generator
import acoustics.imaging
import acoustics.octave
import acoustics.power
import acoustics.quantity
import acoustics.reflection
import acoustics.room
import acoustics.signal
#import acoustics.utils
import acoustics.weighting

from acoustics._signal import Signal



"""
ISO 1996-1:2003
===============

ISO 1996-1:2003 defines the basic quantities to be used for the description of 
noise in community environments and describes basic assessment procedures. It 
also specifies methods to assess environmental noise and gives guidance on 
predicting the potential annoyance response of a community to long-term exposure 
from various types of environmental noises. The sound sources can be separate or 
in various combinations. Application of the method to predict annoyance response 
is limited to areas where people reside and to related long-term land uses.


"""
import numpy as np

def composite_rating_level(levels, hours, adjustment):
    """Composite rating level.
    
    :params levels: Level per period.
    :params hours: Amount of hours per period.
    :params adjustment: Adjustment per period.
    
    Composite whole-day rating levels are calculated as
    
    .. math:: L_R = 10 \\log{\\left[ \\sum_i \\frac{d_i}{24} 10^{(L_i+K_i)/10}  \\right]}
    
    where :math:`i` is a period. See equation 6 and 7 of the standard.
    
    .. note:: Summation is done over the last axis.
    """
    levels = np.asarray(levels)
    hours = np.asarray(hours)
    adjustment = np.asarray(adjustment)
    
    return 10.0 * np.log10( (hours/24.0 * 10.0**((levels+adjustment)/10.0)).sum(axis=-1))



"""
Room
====

The room acoustics module contains several functions to calculate the reverberation time in spaces.
"""

from __future__ import division

import numpy as np

from scipy.io import wavfile
from scipy import stats

from acoustics.utils import _is_1d
from acoustics.signal import bandpass
from acoustics.bands import (_check_band_type, octave_low, octave_high,
                                  third_low, third_high)


SOUNDSPEED = 343.0

def mean_alpha(alphas, surfaces):
    """
    Calculate mean of absorption coefficients.
    
    :param alphas: Absorption coefficients :math:`\\alpha`.
    :param surfaces: Surfaces :math:`S`.
    """
    return np.average(alphas, axis=0, weights=surfaces)


def nrc(alphas):
    """
    Calculate Noise Reduction Coefficient (NRC) from four absorption
    coefficient values (250, 500, 1000 and 2000 Hz).
    
    :param alphas: Absorption coefficients :math:`\\alpha`.
    
    """
    alpha_axis = alphas.ndim - 1
    return np.mean(alphas, axis=alpha_axis)


def t60_sabine(surfaces, alpha, volume, c=SOUNDSPEED):
    """
    Reverberation time according to Sabine.
    
    :param surfaces: Surface of the room :math:`S`. 
        NumPy array that contains different surfaces.
    :type surfaces: :class:`np.ndarray`
    :param alpha: Absorption coefficient of the room :math:`\\alpha`.
        Contains absorption coefficients of ``surfaces``.
        It could be one value or some values in different bands (1D and 2D
        array, respectively).
    :type alpha: :class:`np.ndarray`
    :param volume: Volume of the room :math:`V`.
    :type volume: :class:`float`
    :param c: Speed of sound :math:`c`.
    :type c: :class:`float`
    :returns: Reverberation time :math:`T_{60}`
    
    Sabine's formula for the reverberation time is:
    
    .. math:: T_{60} = \\frac{24 \\ln(10)}{c} \\frac{V}{S\\alpha}
    
    """
    mean_alpha = np.average(alpha, axis=0, weights=surfaces)
    S = np.sum(surfaces, axis=0)
    A = S * mean_alpha
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_eyring(surfaces, alpha, volume, c=SOUNDSPEED):
    """
    Reverberation time according to Eyring.
    
    :param surfaces: Surfaces :math:`S`.
    :param alpha: Mean absorption coefficient :math:`\\alpha` or by frequency bands
    :param volume: Volume of the room :math:`V`.
    :param c: Speed of sound :math:`c`.
    :returns: Reverberation time :math:`T_{60}`
    
    Eyring's formula for the reverberation time is:
    
    .. math:: T_{60} = \\frac{24 \\ln{10} V}{c \\left( 4 mV - S \\ln{\\left( 1 - \\alpha \\right)} \\right)}

    """
    mean_alpha = np.average(alpha, axis=0, weights=surfaces)
    S = np.sum(surfaces, axis=0)
    A = -S * np.log(1-mean_alpha)
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_millington(surfaces, alpha, volume, c=SOUNDSPEED):
    """
    Reverberation time according to Millington.
    
    :param surfaces: Surfaces :math:`S`.
    :param alpha: Mean absorption coefficient :math:`\\alpha` or by frequency bands
    :param volume: Volume of the room :math:`V`.
    :param c: Speed of sound :math:`c`.
    :returns: Reverberation time :math:`T_{60}`
    """
    mean_alpha = np.average(alpha, axis=0, weights=surfaces)
    A = -np.sum(surfaces[:, np.newaxis] * np.log(1.0 - mean_alpha), axis=0)
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_fitzroy(surfaces, alpha, volume, c=SOUNDSPEED):
    """
    Reverberation time according to Fitzroy.
    
    :param surfaces: Surfaces :math:`S`.
    :param alpha: Mean absorption coefficient :math:`\\alpha` or by frequency bands
    :param volume: Volume of the room :math:`V`.
    :param c: Speed of sound :math:`c`.
    :returns: Reverberation time :math:`T_{60}`
    """
    Sx = np.sum(surfaces[0:2])
    Sy = np.sum(surfaces[2:4])
    Sz = np.sum(surfaces[4:6])
    St = np.sum(surfaces)
    alpha = _is_1d(alpha)
    a_x = np.average(alpha[:, 0:2], weights=surfaces[0:2], axis=1)
    a_y = np.average(alpha[:, 2:4], weights=surfaces[2:4], axis=1)
    a_z = np.average(alpha[:, 4:6], weights=surfaces[4:6], axis=1)
    factor = -(Sx / np.log(1.0-a_x) + Sy / np.log(1.0-a_y) + Sz / np.log(1-a_z))
    t60 = 4.0 * np.log(10.0**6.0) * volume * factor / (c * St**2.0)
    return t60


def t60_arau(Sx, Sy, Sz, alpha, volume, c=SOUNDSPEED):
    """
    Reverberation time according to Arau. [#arau]_

    :param Sx: Total surface perpendicular to x-axis (yz-plane) :math:`S_{x}`.
    :param Sy: Total surface perpendicular to y-axis (xz-plane) :math:`S_{y}`.
    :param Sz: Total surface perpendicular to z-axis (xy-plane) :math:`S_{z}`.
    :param alpha: Absorption coefficients :math:`\\mathbf{\\alpha} = \\left[ \\alpha_x, \\alpha_y, \\alpha_z \\right]`
    :param volume: Volume of the room :math:`V`.
    :param c: Speed of sound :math:`c`.
    :returns: Reverberation time :math:`T_{60}`
    
    .. [#arau] For more details, please see
       http://www.arauacustica.com/files/publicaciones/pdf_esp_7.pdf
    """
    a_x = -np.log(1 - alpha[0])
    a_y = -np.log(1 - alpha[1])
    a_z = -np.log(1 - alpha[2])
    St = np.sum(np.array([Sx, Sy, Sz]))
    A = St * a_x**(Sx/St) * a_y**(Sy/St) * a_z**(Sz/St)
    t60 = 4.0 * np.log(10.0**6.0) * volume / (c * A)
    return t60


def t60_impulse(file_name, bands, rt='t30'):
    """
    Reverberation time from a WAV impulse response.

    :param file_name: name of the WAV file containing the impulse response.
    :param bands: Octave or third bands as NumPy array.
    :param rt: Reverberation time estimator. It accepts `'t30'`, `'t20'`, `'t10'` and `'edt'`.
    :returns: Reverberation time :math:`T_{60}`
    
    """
    fs, raw_signal = wavfile.read(file_name)
    band_type = _check_band_type(bands)

    if band_type is 'octave':
        low = octave_low(bands[0], bands[-1])
        high = octave_high(bands[0], bands[-1])
    elif band_type is 'third':
        low = third_low(bands[0], bands[-1])
        high = third_high(bands[0], bands[-1])

    rt = rt.lower()
    if rt == 't30':
        init = -5.0
        end = -35.0
        factor = 2.0
    elif rt == 't20':
        init = -5.0
        end = -25.0
        factor = 3.0
    elif rt == 't10':
        init = -5.0
        end = -15.0
        factor = 6.0
    elif rt == 'edt':
        init = 0.0
        end = -10.0
        factor = 6.0

    t60 = np.zeros(bands.size)

    for band in range(bands.size):
        # Filtering signal
        filtered_signal = bandpass(raw_signal, low[band],
                                                 high[band], fs, order=8)
        abs_signal = np.abs(filtered_signal) / np.max(np.abs(filtered_signal))

        # Schroeder integration
        sch = np.cumsum(abs_signal[::-1]**2)[::-1]
        sch_db = 10.0 * np.log10(sch / np.max(sch))
        
        # Linear regression
        sch_init = sch_db[np.abs(sch_db - init).argmin()]
        sch_end = sch_db[np.abs(sch_db - end).argmin()]
        init_sample = np.where(sch_db == sch_init)[0][0]
        end_sample = np.where(sch_db == sch_end)[0][0]
        x = np.arange(init_sample, end_sample + 1) / fs
        y = sch_db[init_sample: end_sample + 1]
        slope, intercept, r_value, p_value, std_err = stats.linregress(x,y)

        # Reverberation time (T30, T20, T10 or EDT)
        db_regress_init = (init - intercept) / slope
        db_regress_end = (end - intercept) / slope
        t60[band] = factor * (db_regress_end - db_regress_init)
    return t60


def clarity(time, signal, fs, bands=None):
    """
    Clarity :math:`C_i` determined from an impulse response.
    
    :param time: Time in miliseconds (e.g.: 50, 80).
    :param signal: Impulse response.
    :type signal: :class:`np.ndarray`
    :param fs: Sample frequency.
    :param bands: Bands of calculation (optional). Only support standard octave and third-octave bands.
    :type bands: :class:`np.ndarray`
    
    """
    band_type = _check_band_type(bands)

    if band_type is 'octave':
        low = octave_low(bands[0], bands[-1])
        high = octave_high(bands[0], bands[-1])
    elif band_type is 'third':
        low = third_low(bands[0], bands[-1])
        high = third_high(bands[0], bands[-1])

    c = np.zeros(bands.size)
    for band in range(bands.size):
        filtered_signal = bandpass(signal, low[band],high[band],
                                                 fs, order=8)
        h2 = filtered_signal**2.0
        t = int((time/1000.0)*fs + 1)
        c[band] = 10.0*np.log10((np.sum(h2[:t])/np.sum(h2[t:])))
    return c


def c50_from_file(file_name, bands=None):
    """
    Clarity for 50 miliseconds :math:`C_{50}` from a file.
    
    :param file_name: File name (only WAV is supported).
    :type file_name: :class:`str`
    :param bands: Bands of calculation (optional). Only support standard octave and third-octave bands.
    :type bands: :class:`np.ndarray`
    
    """
    fs, signal = wavfile.read(file_name)
    return clarity(50.0, signal, fs, bands)


def c80_from_file(file_name, bands=None):
    """
    Clarity for 80 miliseconds :math:`C_{80}` from a file.
    
    :param file_name: File name (only WAV is supported).
    :type file_name: :class:`str`
    :param bands: Bands of calculation (optional). Only support standard octave and third-octave bands.
    :type bands: :class:`np.ndarray`
    
    """
    fs, signal = wavfile.read(file_name)
    return clarity(80.0, signal, fs, bands)



"""
ISO 9613-1:1993
===============

ISO 9613-1:1993 specifies an analytical method of calculating the attenuation of sound 
as a result of atmospheric absorption for a variety of meteorological conditions.

"""

import numpy as np


SOUNDSPEED = 343.2
"""
Speed of sound.
"""


REFERENCE_TEMPERATURE = 293.15
"""
Reference temperature.
"""

REFERENCE_PRESSURE = 101.325
"""
International Standard Atmosphere in kilopascal.
"""

TRIPLE_TEMPERATURE = 273.16
""".
Triple point isotherm temperature.
"""

def soundspeed(temperature, reference_temperature=REFERENCE_TEMPERATURE):
    """
    Speed of sound :math:`c`.
    
    :param temperature: Ambient temperature :math:`T_0`
    :param reference_temperature: Reference temperature :math:`T`
    
    
    The speed of sound is calculated using
    
    ..  math:: c = 343.2 \\left( \\frac{T}{T_0} \\right)
    
    """
    return 343.2 * np.sqrt(temperature / reference_temperature)
    
def saturation_pressure(temperature, reference_pressure=REFERENCE_PRESSURE, triple_temperature=TRIPLE_TEMPERATURE):
    """
    Saturation vapour pressure :math:`p_{sat}`.
    
    :param temperature: Ambient temperature :math:`T`
    :param reference_pressure: Reference pressure :math:`p_r`
    :param triple_temperature: Triple point temperature water :math:`T_{01}`
    
    The saturation vapour pressure is calculated using
    
    .. math:: p_{sat} = 10^C \cdot p_r
    
    with exponent :math:`C` given by
    
    .. math:: C = -6.8346 \cdot \\left( \\frac{T_{01}}{T} \\right)^{1.261}  + 4.6151
    
    """
    return reference_pressure * 10.0** (-6.8346 *(triple_temperature/temperature)**(1.261) + 4.6151)

def molar_concentration_water_vapour(relative_humidity, saturation_pressure, pressure):
    """
    Molar concentration of water vapour :math:`h`.
    
    :param relative_humidity: Relative humidity :math:`h_r`
    :param saturation_pressure: Saturation pressure :math:`p_{sat}`
    :param pressure: Ambient pressure :math:`p`
   
    The molar concentration of water vapour is calculated using
   
    .. math:: h = h_r  \\frac{p_{sat}}{p_a}
    
    """
    return relative_humidity * saturation_pressure / pressure

def relaxation_frequency_oxygen(pressure, h, reference_pressure=REFERENCE_PRESSURE):
    """
    Relaxation frequency of oxygen :math:`f_{r,O}`.
    
    :param pressure: Ambient pressure :math:`p_a`
    :param reference_pressure: Reference pressure :math:`p_r`
    :param h: Molar concentration of water vapour :math:`h`
    
    The relaxation frequency of oxygen is calculated using
    
    .. math:: f_{r,O} = \\frac{p_a}{p_r} \\left( 24 + 4.04 \cdot 10^4 h \\frac{0.02 + h}{0.391 + h}  \\right)
    
    """
    return pressure / reference_pressure * ( 24.0 + 4.04 * 10.0**4.0 * h * (0.02 + h) / (0.391 + h) )

def relaxation_frequency_nitrogen(pressure, temperature, h, reference_pressure=REFERENCE_PRESSURE, reference_temperature=REFERENCE_TEMPERATURE):        
    """
    Relaxation frequency of nitrogen :math:`f_{r,N}`.
    
    :param pressure: Ambient pressure :math:`p_a`
    :param temperature: Ambient temperature :math:`T`
    :param h: Molar concentration of water vapour :math:`h`
    :param reference_pressure: Reference pressure :math:`p_{ref}`
    :param reference_temperature: Reference temperature :math:`T_{ref}`
    
    The relaxation frequency of nitrogen is calculated using
    
    .. math:: f_{r,N} = \\frac{p_a}{p_r} \\left( \\frac{T}{T_0} \\right)^{-1/2} \cdot \\left( 9 + 280 h \exp{\\left\{ -4.170 \\left[ \\left(\\frac{T}{T_0} \\right)^{-1/3} -1 \\right] \\right\} } \\right)
    
    """
    return pressure / reference_pressure * (temperature/reference_temperature)**(-0.5) * (9.0 + 280.0 * h * np.exp(-4.170 * ((temperature/reference_temperature)**(-1.0/3.0) - 1.0 ) ) )


def attenuation_coefficient(pressure, temperature, reference_pressure, reference_temperature, relaxation_frequency_nitrogen, relaxation_frequency_oxygen, frequency):
    """
    Attenuation coefficient :math:`\\alpha` describing atmospheric absorption in dB/m for the specified ``frequency``.
    
    :param pressure: Ambient pressure :math:`T`
    :param temperature: Ambient temperature :math:`T`
    :param reference_pressure: Reference pressure :math:`p_{ref}`
    :param reference_temperature: Reference temperature :math:`T_{ref}`
    :param relaxation_frequency_nitrogen: Relaxation frequency of nitrogen :math:`f_{r,N}`.
    :param relaxation_frequency_oxygen: Relaxation frequency of oxygen :math:`f_{r,O}`.
    :param frequency: Frequencies to calculate :math:`\\alpha` for.
    
    """
    return 8.686 * frequency**2.0 * ( ( 1.84 * 10.0**(-11.0) * (reference_pressure/pressure) * (temperature/reference_temperature)**(0.5)) + (temperature/reference_temperature)**(-2.5) * ( 0.01275 * np.exp(-2239.1 / temperature) * (relaxation_frequency_oxygen + (frequency**2.0/relaxation_frequency_oxygen))**(-1.0) + 0.1068 * np.exp(-3352.0/temperature) * (relaxation_frequency_nitrogen + (frequency**2.0/relaxation_frequency_nitrogen))**(-1.0) ) )



"""
Utils
=====

"""

from __future__ import division

import numpy as np

SOUNDSPEED = 343.0
"""
Speed of sound in air.
"""

from acoustics.decibel import dbsum

esum = dbsum

def mean_tl(tl, surfaces):
    try:
        tau_axis = tl.ndim - 1
    except AttributeError:
        tau_axis = 0
    tau = 1.0 / (10.0**(tl/10.0))
    return 10.0 * np.log10(1.0 / np.average(tau, tau_axis, surfaces))


def wavelength(freq, c=SOUNDSPEED):
    """
    Wavelength for one or more frequencies (as ``NumPy array``).
    """
    return c/freq


def w(freq):
    """
    Angular frequency for one o more frequencies (as ``NumPy array``).
    """
    return 2.0 * np.pi * freq


def _is_1d(input):
    if type(input) is int or type(input) is float:
        return input
    elif input.ndim == 1:
        return np.array([input])
    elif input.ndim == 2 and input.shape[0] == 1:
        return input[0]
    else:
        return input



"""
Atmosphere
==========

The atmosphere module contains functions and classes related to atmospheric 
acoustics and is based on :mod:`acoustics.standards.iso_9613_1_1993`.

Atmosphere class
****************

.. autoclass:: acoustics.atmosphere.Atmosphere

From ISO 9613-1 1993
********************

Constants
---------

.. autoattribute:: acoustics.standards.iso_9613_1_1993.SOUNDSPEED
.. autoattribute:: acoustics.standards.iso_9613_1_1993.REFERENCE_TEMPERATURE
.. autoattribute:: acoustics.standards.iso_9613_1_1993.REFERENCE_PRESSURE
.. autoattribute:: acoustics.standards.iso_9613_1_1993.TRIPLE_TEMPERATURE

Functions
---------

.. autofunction:: acoustics.standards.iso_9613_1_1993.soundspeed
.. autofunction:: acoustics.standards.iso_9613_1_1993.saturation_pressure
.. autofunction:: acoustics.standards.iso_9613_1_1993.molar_concentration_water_vapour
.. autofunction:: acoustics.standards.iso_9613_1_1993.relaxation_frequency_nitrogen
.. autofunction:: acoustics.standards.iso_9613_1_1993.relaxation_frequency_oxygen
.. autofunction:: acoustics.standards.iso_9613_1_1993.attenuation_coefficient




"""
from __future__ import division, print_function

import numpy as np
import matplotlib.pyplot as plt

import acoustics
from acoustics.standards.iso_9613_1_1993 import *
                                                

class Atmosphere(object):
    """
    Class describing atmospheric conditions.
    """
    
    REF_TEMP = 293.15
    """Reference temperature"""
    
    REF_PRESSURE = 101.325
    """International Standard Atmosphere in kilopascal"""
    
    TRIPLE_TEMP = 273.16
    """Triple point isotherm temperature."""

    def __init__(self, 
                 temperature=REFERENCE_TEMPERATURE, 
                 pressure=REFERENCE_PRESSURE,
                 relative_humidity=0.0, 
                 reference_temperature=REFERENCE_TEMPERATURE,
                 reference_pressure=REFERENCE_PRESSURE,
                 triple_temperature=TRIPLE_TEMPERATURE):
        """

        :param temperature: Temperature in kelvin
        :param pressure: Pressure
        :param relative_humidity: Relative humidity
        :param reference_temperature: Reference temperature.
        :param reference_pressure: Reference pressure.
        :param triple_temperature: Triple temperature.
        """
        
        self.temperature = temperature
        """Ambient temperature :math:`T`."""
        
        self.pressure = pressure
        """Ambient pressure :math:`p_a`."""
        
        self.relative_humidity = relative_humidity
        """Relative humidity"""
        
        self.reference_temperature = reference_temperature
        """
        Reference temperature.
        """
        
        self.reference_pressure = reference_pressure
        """
        Reference pressure.
        """
        
        self.triple_temperature = triple_temperature
        """
        Triple temperature.
        """
    
    def __repr__(self):
        return "Atmosphere{}".format(self.__str__())
    
    def __str__(self):
        return "(temperature={}, pressure={}, relative_humidity={}, " \
               "reference_temperature={}, reference_pressure={}, " \
               "triple_temperature={})".format(self.temperature, self.pressure, 
                                            self.relative_humidity, 
                                            self.reference_temperature, 
                                            self.reference_pressure, 
                                            self.triple_temperature)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__ and self.__class__ == other.__class__

    @property
    def soundspeed(self):
        """
        Speed of sound :math:`c`. 
        
        The speed of sound is calculated using :func:`acoustics.standards.iso_9613_1_1993.soundspeed`.
        """
        return soundspeed(self.temperature, self.reference_temperature)
        
    @property
    def saturation_pressure(self):
        """
        Saturation pressure :math:`p_{sat}`.
        
        The saturation pressure is calculated using :func:`acoustics.standards.iso_9613_1_1993.saturation_pressure`.
        """
        return saturation_pressure(self.temperature, self.reference_pressure, self.triple_temperature)
    
    @property
    def molar_concentration_water_vapour(self):
        """
        Molar concentration of water vapour :math:`h`.
        
        The molar concentration of water vapour is calculated using :func:`acoustics.standards.iso_9613_1_1993.molar_concentration_water_vapour`.
        """
        return molar_concentration_water_vapour(self.relative_humidity, self.saturation_pressure, self.pressure)
        
    @property
    def relaxation_frequency_nitrogen(self):
        """
        Resonance frequency of nitrogen :math:`f_{r,N}`.
        
        The resonance frequency is calculated using :func:`acoustics.standards.iso_9613_1_1993.relaxation_frequency_nitrogen`.
        """
        return relaxation_frequency_nitrogen(self.pressure, self.temperature, self.molar_concentration_water_vapour, self.reference_pressure, self.reference_temperature)
    
    @property
    def relaxation_frequency_oxygen(self):
        """
        Resonance frequency of oxygen :math:`f_{r,O}`.
        
        The resonance frequency is calculated using :func:`acoustics.standards.iso_9613_1_1993.relaxation_frequency_oxygen`.
        """
        return relaxation_frequency_oxygen(self.pressure, self.molar_concentration_water_vapour, self.reference_pressure)
    
    def attenuation_coefficient(self, frequency):
        """
        Attenuation coefficient :math:`\\alpha` describing atmospheric absorption in dB/m as function of ``frequency``.
        
        :param frequency: Frequencies to be considered.
        
        The attenuation coefficient is calculated using :func:`acoustics.standards.iso_9613_1_1993.attenuation_coefficient`.
        """
        return attenuation_coefficient(self.pressure, self.temperature, self.reference_pressure, self.reference_temperature, self.relaxation_frequency_nitrogen, self.relaxation_frequency_oxygen, frequency)

    def frequency_response(self, distance, frequencies, inverse=False):
        """Frequency response.

        :param distance: Distance between source and receiver.
        :param frequencies: Frequencies for which to compute the response.
        :param inverse: Whether the attenuation should be undone.

        """
        return frequency_response(self, distance, frequencies, inverse)

    def impulse_response(self, distance, fs, ntaps=None, inverse=False):
        """Impulse response of sound travelling through `atmosphere` for a given `distance` sampled at `fs`.

        :param atmosphere: Atmosphere.
        :param distance: Distance between source and receiver.
        :param fs: Sample frequency
        :param ntaps: Amount of taps.
        :param inverse: Whether the attenuation should be undone.

        .. seealso:: :func:`impulse_response`
        """
        return impulse_response(self, distance, fs, ntaps, inverse)

    def plot_attenuation_coefficient(self, frequency):
        """
        Plot the attenuation coefficient :math:`\\alpha` as function of frequency and write the figure to ``filename``.

        :param filename: Filename
        :param frequency: Frequencies

        .. note:: The attenuation coefficient is plotted in dB/km!

        """
        fig = plt.figure()
        ax0 = fig.add_subplot(111)
        ax0.plot(frequency, self.attenuation_coefficient(frequency)*1000.0)
        ax0.set_xscale('log')
        ax0.set_yscale('log')
        ax0.set_xlabel(r'$f$ in Hz')
        ax0.set_ylabel(r'$\alpha$ in dB/km')
        ax0.legend()

        return fig


def frequency_response(atmosphere, distance, frequencies, inverse=False):
    """Single-sided frequency response.

    :param atmosphere: Atmosphere.
    :param distance: Distance between source and receiver.
    :param frequencies: Frequencies for which to compute the response.
    :param inverse: Whether the attenuation should be undone.
    """
    sign = +1 if inverse else -1
    tf = 10.0**( float(sign) * distance * atmosphere.attenuation_coefficient(frequencies) / 20.0  )
    return tf

def impulse_response(atmosphere, distance, fs, ntaps, inverse=False):
    """Impulse response of sound travelling through `atmosphere` for a given `distance` sampled at `fs`.

    :param atmosphere: Atmosphere.
    :param distance: Distance between source and receiver.
    :param fs: Sample frequency
    :param ntaps: Amount of taps.
    :param inverse: Whether the attenuation should be undone.

    The attenuation is calculated for a set of positive frequencies. Because the attenuation is the same for the negative frequencies, we have Hermitian symmetry.
    The attenuation is entirely real-valued. We like to have a constant group delay and therefore we need a linear-phase filter.
    This function creates a zero-phase filter, which is the special case of a linear-phase filter with zero phase slope.
    The type of filter is non-causal.
    The impulse response of the filter is made causal by rotating it by M/2 samples and discarding the imaginary parts.
    A real, even impulse response corresponds to a real, even frequency response.
    """
    # Frequencies vector with positive frequencies only.
    frequencies = np.fft.rfftfreq(ntaps, 1./fs)
    # Single-sided spectrum. Negative frequencies have the same values.
    tf = frequency_response(atmosphere, distance, frequencies, inverse)
    # Impulse response. We design a zero-phase filter (linear-phase with zero slope).
    # We need to shift the IR to make it even. Taking the real part should not be necessary, see above.
    #ir = np.fft.ifftshift(np.fft.irfft(tf, n=ntaps)).real
    ir = acoustics.signal.impulse_response_real_even(tf, ntaps=ntaps)
    return ir


__all__ = ['Atmosphere', 'SOUNDSPEED', 'REFERENCE_TEMPERATURE', 
           'REFERENCE_TEMPERATURE', 'TRIPLE_TEMPERATURE',
           'soundspeed', 'saturation_pressure', 
           'molar_concentration_water_vapour', 
           'relaxation_frequency_oxygen', 
           'relaxation_frequency_nitrogen', 
           'attenuation_coefficient',
           'impulse_response',
           'frequency_response'
           ]



"""
Decibel
=======

The `decibel` module contains basic functions for decibel arithmetic.

"""

import numpy as np

def dbsum(levels, axis=None):
    """Energetic summation of levels.
    
    :param levels: Sequence of levels.
    :param axis: Axis over which to perform the operation.
    
    .. math:: L_{sum} = 10 \\log_{10}{\\sum_{i=0}^n{10^{L/10}}}
    
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0**(levels/10.0)).sum(axis=axis))


def dbmean(levels, axis=None):
    """Energetic average of levels.
    
    :param levels: Sequence of levels.
    :param axis: Axis over which to perform the operation.
    
    .. math:: L_{mean} = 10 \\log_{10}{\\frac{1}{n}\\sum_{i=0}^n{10^{L/10}}}
    
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0**(levels/10.0)).mean(axis=axis))
    

def dbadd(a, b):
    """Energetic addition.
    
    :param a: Single level or sequence of levels.
    :param b: Single level or sequence of levels.
    
    .. math:: L_{a+b} = 10 \\log_{10}{10^{L_b/10}+10^{L_a/10}}
    
    Energetically adds b to a.
    """
    a = np.asanyarray(a)
    b = np.asanyarray(b)
    return 10.0*np.log10(10.0**(a/10.0)+10.0**(b/10.0))


def dbsub(a, b):
    """Energetic subtraction.
    
    :param a: Single level or sequence of levels.
    :param b: Single level or sequence of levels.
    
    .. math:: L_{a-b} = 10 \\log_{10}{10^{L_a/10}-10^{L_b/10}}
    
    Energitally subtract b from a.
    """
    a = np.asanyarray(a)
    b = np.asanyarray(b)
    return 10.0*np.log10(10.0**(a/10.0)-10.0**(b/10.0))


def dbmul(levels, f, axis=None):
    """Energetically add `levels` `f` times.
    
    :param levels: Sequence of levels.
    :param f: Multiplication factor `f`.
    :param axis: Axis over which to perform the operation.
    
    .. math:: L_{sum} = 10 \\log_{10}{\\sum_{i=0}^n{10^{L/10} \cdot f}}
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0**(levels/10.0) * f).sum(axis=axis))


def dbdiv(levels, f, axis=None):
    """Energetically divide `levels` `f` times.
    
    :param levels: Sequence of levels.
    :param f: Divider `f`.
    :param axis: Axis over which to perform the operation.
    
    .. math:: L_{sum} = 10 \\log_{10}{\\sum_{i=0}^n{10^{L/10} / f}}
    
    """
    levels = np.asanyarray(levels)
    return 10.0 * np.log10((10.0**(levels/10.0) / f).sum(axis=axis))

__all__ = ['dbsum', 'dbmean', 'dbadd', 'dbsub', 'dbmul', 'dbdiv']



"""
Weighting
=========

Weightings according to IEC 61672-1:2003.


"""
from __future__ import division

import numpy as np

from acoustics.bands import third

THIRD_OCTAVE_A_WEIGHTING = np.array([-63.4, -56.7, -50.5, -44.7, -39.4, -34.6, 
                                     -30.2, -26.2, -22.5, -19.1, -16.1, -13.4, 
                                     -10.9, -8.6, -6.6, -4.8, -3.2, -1.9, 
                                     -0.8, +0.0, +0.6, +1.0, +1.2, +1.3, 
                                     +1.2, +1.0, +0.5, -0.1, -1.1, -2.5, 
                                     -4.3, -6.6, -9.3])
"""
A-weighting filter for preferred 1/3-octave band center frequencies, as specified in :attr:`acoustics.bands.THIRD_OCTAVE_CENTER_FREQUENCIES`.
"""

THIRD_OCTAVE_C_WEIGHTING = np.array([-11.2, -8.5, -6.2, -4.4, -3.0, -2.0, 
                                     -1.3, -0.8, -0.5, -0.3, -0.2, -0.1, 
                                     +0.0, +0.0, +0.0, +0.0, +0.0, +0.0,
                                     +0.0, +0.0, +0.0, -0.1, -0.2, -0.3,
                                     -0.5, -0.8, -1.3, -2.0, -3.0, -4.4,
                                     -6.2, -8.5, -11.2])
"""
C-weighting filter for preferred 1/3-octave band center frequencies, as specified in :attr:`acoustics.bands.THIRD_OCTAVE_CENTER_FREQUENCIES`.
"""


def a_weighting(first, last):
    """
    Select frequency weightings between ``first`` and ``last``
    centerfrequencies from A-weighting.
    Possible values for these frequencies are third-octave frequencies
    between 12.5 Hz and 20,000 Hz (including them).

    Parameters
    ----------
    first : scalar
       First third-octave centerfrequency.

    last : scalar
        Last third-octave centerfrequency.

    Returns
    -------
    NumPy array with A-weighting between ``first`` and ``last``
    centerfrequencies.
    """
    return _weighting("a", first, last)


def c_weighting(first, last):
    """
    Select frequency weightings between ``first`` and ``last``
    centerfrequencies from C-weighting.
    Possible values for these frequencies are third-octave frequencies
    between 12.5 Hz and 20,000 Hz (including them).

    Parameters
    ----------
    first : scalar
       First third-octave centerfrequency.

    last : scalar
        Last third-octave centerfrequency.

    Returns
    -------
    NumPy array with A-weighting between ``first`` and ``last``
    centerfrequencies.
    """
    return _weighting("c", first, last)


def _weighting(filter_type, first, last):
    third_oct_bands = third(12.5, 20000.0)
    low = np.where(third_oct_bands == first)[0]
    high = np.where(third_oct_bands == last)[0]

    if filter_type == "a":
        freq_weightings = THIRD_OCTAVE_A_WEIGHTING

    elif filter_type == "c":
        freq_weightings = THIRD_OCTAVE_C_WEIGHTING
        
    return freq_weightings[low: high+1]


def z2a(levels, first, last):
    """Apply A-weighting to Z-weighted signal.
    """
    return levels + a_weighting(first, last)


def a2z(levels, first, last):
    """Remove A-weighting from A-weighted signal.
    """
    return levels - a_weighting(first, last)


def z2c(levels, first, last):
    """Apply C-weighting to Z-weighted signal.
    """
    return levels + c_weighting(first, last)


def c2z(levels, first, last):
    """Remove C-weighting from C-weighted signal.
    """
    return levels - c_weighting(first, last)


def a2c(levels, first, last):
    """Go from A-weighted to C-weighted.
    """
    dB = a2z(levels, first, last)
    return z2c(dB, first, last)


def c2a(levels, first, last):
    """Go from C-weighted to A-weighted.
    """
    dB = c2z(levels, first, last)
    return z2a(dB, first, last)

__all__ = ['THIRD_OCTAVE_A_WEIGHTING', 
           'THIRD_OCTAVE_C_WEIGHTING', 
           'a_weighting',
           'c_weighting',
           'z2a',
           'a2z',
           'z2c',
           'c2z',
           'a2c',
           'c2a',
           ]



"""
Doppler
=======

Doppler shift module.
"""
from __future__ import division

SOUNDSPEED = 343.0
"""Speed of sound
"""

def velocity_from_doppler_shift(f1, f2, c=SOUNDSPEED):
    """Calculate velocity based on measured frequency shifts due to Doppler shift.
    
    :param c: Speed of sound :math:`c`.
    :param f1: Lower frequency :math:`f_1`.
    :param f2: Upper frequency :math:`f_2`.
    
    .. math:: v = c \cdot \\left( \\frac{f_2 - f_1}{f_2 + f_1} \\right)
    
    The assumption is made that the velocity is constant between the observation times.
    
    """
    return c * (f2 - f1) / (f2 + f1)


def frequency_shift(frequency, velocity_source, velocity_receiver, soundspeed=SOUNDSPEED):
    """Frequency shift due to Doppler effect.
    
    :param frequency: Emitted frequency :math:`f`.
    :param velocity_source: Velocity of source :math:`v_s`. Positive if the source is moving away from the receiver (and negative in the other direction).
    :param velocity_receiver: Velocity of receiver :math:`v_r`. Positive if the receiver is moving towards the source (and negative in the other direction);
    :param soundspeed: Speed of sound :math:`c`.
    
    .. math:: f = \\frac{c + v_r}{c + v_s} f_0
    
    """
    return (soundspeed + velocity_receiver) / (soundspeed + velocity_source) * frequency



import os
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

import numpy as np
from Cython.Build import cythonize

if os.path.exists('README.md'):
    long_description = open('README.md').read()
else:
    long_description = "A Python library aimed at acousticians."


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

setup(
      name='acoustics',
      version='0.1.2',
      description="Acoustics module for Python.",
      long_description=long_description,
      author='Python Acoustics',
      author_email='fridh@fridh.nl',
      license='LICENSE',
      packages=find_packages(exclude=["tests"]),
      package_dir={'acoustics': 'acoustics'},
      package_data={'acoustics': ['data/*']},
      #py_modules=['turbulence'],
      scripts=[],
      zip_safe=False,
      install_requires=[
          'numpy >=1.8',
          'scipy >= 0.16',
          'matplotlib',
          'six >= 1.4.1',
          'cython',
          'pandas>=0.15',
          'tabulate',
          ],
      extras_require={
          'documentation': 'sphinx',
          'jit': 'numba',
          'fast_fft': 'pyFFTW',
          },
      tests_require=['pytest'],
      cmdclass={'test': PyTest},
      ext_modules=cythonize('acoustics/*.pyx'),
      include_dirs=[np.get_include()]
      )



"""
Directivity
===========

The directivity module provides tools to work with directivity.

The following conventions are used within this module:

* The inclination angle :math:`\\theta` has a range :math:`[0, \\pi]`.
* The azimuth angle :math:`\\phi` has a range :math:`[0 , 2 \\pi]`.

"""
from __future__ import division

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.cm as cm
from matplotlib.colors import Normalize
import numpy as np
import abc
from scipy.interpolate import interp2d as interpolate
from scipy.special import sph_harm
    
    
def cardioid(theta, a=1.0, k=1.0):
    """
    A cardioid pattern.
    
    :param a: a
    :param k: k
    """
    return np.abs( a + a * np.cos(k*theta) )

def figure_eight(theta, phi=0.0):
    """
    A figure-of-eight pattern.
    
    :param theta: angle :math:`\\theta`
    """
    #return spherical_harmonic(theta, phi, m=0, n=1)
    return np.abs( np.cos(theta) )

def spherical_harmonic(theta, phi, m=0, n=0):
    """Spherical harmonic of order `m` and degree `n`.
    
    .. note:: The degree `n` is often denoted `l`.
    
    .. seealso:: :func:`scipy.special.sph_harm`
    
    """
    return sph_harm(m, n, phi, theta).real
    

def spherical_to_cartesian(r, theta , phi):
    """
    Convert spherical coordinates to cartesian coordinates.
    
    :param r: norm
    :param theta: angle :math:`\\theta`
    :param phi: angle :math:`\\phi`
    
    .. math:: x = r \\sin{\\theta}\\cos{\\phi}
    .. math:: y = r \\sin{\\theta}\\sin{\\phi}
    .. math:: z = r \\cos{\\theta}
    """
    r = np.asanyarray(r)
    theta = np.asanyarray(theta)
    phi = np.asanyarray(phi)
    return (
        r * np.sin(theta) * np.cos(phi),
        r * np.sin(theta) * np.sin(phi),
        r * np.cos(theta)
        )
    
def cartesian_to_spherical(x, y, z):
    """
    Convert cartesian coordinates to spherical coordinates.
    
    :param x: x
    :param y: y
    :param z: z
    
    .. math:: r = \\sqrt{\\left( x^2 + y^2 + z^2 \\right)}
    .. math:: \\theta = \\arccos{\\frac{z}{r}}
    .. math:: \\phi = \\arccos{\\frac{y}{x}}
    """
    x = np.asanyarray(x)
    y = np.asanyarray(y)
    z = np.asanyarray(z)
    r = np.linalg.norm(np.vstack((x,y,z)), axis=0)
    return r, np.arccos(z/r), np.arctan(y/x)


#def spherical_to_cartesian(spherical):
    #"""
    #Convert spherical coordinates to cartesian coordinates.
    
    #.. math:: x = r \\sin{\\theta}\\cos{\\phi}
    #.. math:: y = r \\sin{\\theta}\\sin{\\phi}
    #.. math:: z = r \\cos{\\theta}
    #"""
    #r = spherical[0,:]
    #theta = spherical[1,:]
    #phi = spherical[2,:]
    #return np.vstack((
        #r * np.sin(theta) * np.cos(phi),
        #r * np.sin(theta) * np.sin(phi),
        #r * np.cos(theta)
        #))
    

#def cartesian_to_spherical(cartesian):
    #"""
    #Convert cartesian coordinates to spherical coordinates.
    
    #.. math:: r = \\sqrt{\\left( x^2 + y^2 + z^2 \\right)}
    #.. math:: \\theta = \\arccos{\\frac{z}{r}}
    #.. math:: \\phi = \\arccos{\\frac{y}{x}}
    #"""
    #x = spherical[0,:]
    #y = spherical[1,:]
    #z = spherical[2,:]
    #r = np.sqrt(np.sum(np.abs(cartesian)**2, axis=0)),
    #return np.vstack((
        #r,
        #np.arccos(z/r),
        #np.arctan(y/x)
        #))


class Directivity(object):
    """
    Abstract directivity class.
    
    This class defines several methods to be implemented by subclasses.
    """
    
    
    def __init__(self, rotation=None):
        
        self.rotation = rotation if rotation else np.array([1.0, 0.0, 0.0]) # X, Y, Z rotation
        """
        Rotation of the directivity pattern.
        """

    @abc.abstractmethod
    def _directivity(self, theta, phi):
        """
        This function should return the directivity as function of :math:`\\theta` and :math:`\\phi`.
        """
        pass

    def _undo_rotation(self, theta, phi):
        """
        Undo rotation.
        """
        pass
        

    def using_spherical(self, r, theta, phi, include_rotation=True):
        """
        Return the directivity for given spherical coordinates.
        
        :param r: norm
        :param theta: angle :math:`\\theta`
        :param phi: angle :math:`\\phi`
        """
        
        """
        Correct for rotation!!!!
        """
        return self._directivity(theta, phi)    
    
    def using_cartesian(self, x, y, z, include_rotation=True):
        """
        Return the directivity for given cartesian coordinates.
        
        :param x: x
        :param y: y
        :param z: z
        """
        return self.using_spherical(*cartesian_to_spherical(x, y, z))
    
    def plot(self, filename=None, include_rotation=True):
        """
        Directivity plot. Plot to ``filename`` when given.
        
        :param filename: Filename
        :param include_rotation: Apply the rotation to the directivity. By default the rotation is applied in this figure.
        """
        
        return plot(self, filename, include_rotation)
        
        
        ##try:
            ##from mayavi import mlab
        ##except ImportError:
            ##raise ImportWarning("mayavi is not available.")
            ##return
        
        #phi = np.linspace(-np.pi, +np.pi, 50)
        #theta = np.linspace(0.0, np.pi, 50)
        
        #theta_n, phi_n = np.meshgrid(theta, phi)    # Create a 2-D mesh
    
        #d = self._directivity(theta_n, phi_n.ravel)
        
        ##fig = plt.figure()
        ##ax0 = fig.add_subplot(111, projection='3d')
        ##ax0.set_title('Directivity')
        
        ##ax0.pcolormesh(u*180.0/np.pi, v*180.0/np.pi, r)
        #(x, y, z) = spherical_to_cartesian(d, theta_n, phi_n)
        
        
        #fig = plt.figure()
        #fig.add_subplot(111, projection='3d')
        #fig.surface()
        
        ##fig = mlab.figure()
        ##s = mlab.mesh(x,y,z)
        ##fig.add(s)
        ##mlab.axes()
        ##mlab.outline()
        ##mlab.show()
        
        ##ax0.plot_wireframe(x*180.0/np.pi, y*180.0/np.pi, z)
        ##ax0.set_xlabel(r'Latitude $u$ in degree')
        ##ax0.set_ylabel(r'Longitude $v$ in degree')
        ##ax0.grid()
        
        ##if filename:
            ##fig.savefig(filename)
    


class Omni(Directivity):
    """
    Class to work with omni-directional directivity.
    """
    
    def _directivity(self, theta, phi):
        """
        Directivity
        """
        return np.ones_like(theta)
    

class Cardioid(Directivity):
    """
    Cardioid directivity.
    """
    
    def _directivity(self, theta, phi):
        """
        Directivity
        """
        return cardioid(theta)
    
class FigureEight(Directivity):
    """Directivity of a figure of eight.
    """
    
    def _directivity(self, theta, phi):
        """Directivity
        """
        return figure_eight(theta, phi)
        

class SphericalHarmonic(Directivity):
    """Directivity of a spherical harmonic of degree `n` and order `m`.
    """
    
    def __init__(self, rotation=None, m=None, n=None):
        
        super().__init__(rotation=rotation)
        self.m = m
        """Order `m`.
        """
        self.n = n
        """Degree `n`.
        """
        
    def _directivity(self, theta, phi):
        """Directivity
        """
        return spherical_harmonic(theta, phi, self.m, self.n)
        
        

class Custom(Directivity):
    """
    A class to work with directivity.
    """
    
    def __init__(self, theta=None, phi=None, r=None):
        """
        Constructor.
        """
        
        self.theta = phi
        """
        Latitude. 1-D array.
        """
        self.phi = phi
        """
        Longitude. 1-D array.
        """
        self.r = r
        """
        Magnitude or radius. 2-D array.
        """
    
    def _directivity(self, theta, phi):
        """
        Custom directivity.
        
        Interpolate the directivity given longitude and latitude vectors.
        """
        f = interpolate(self.theta, self.phi, self.r)
        
        return f(theta, phi)


def plot(d, sphere=False):
    """
    Plot directivity `d`.
    
    :param d: Directivity
    :type d: :class:`Directivity`
    
    :returns: Figure
    """
    
    #phi = np.linspace(-np.pi, +np.pi, 50)
    #theta = np.linspace(0.0, np.pi, 50)
    phi = np.linspace(0.0, +2.0*np.pi, 50)
    theta = np.linspace(0.0, np.pi, 50)
    THETA, PHI = np.meshgrid(theta, phi)
    
    # Directivity strength. Real-valued. Can be positive and negative.
    dr = d.using_spherical(THETA, PHI)
    
    if sphere:
        x, y, z = spherical_to_cartesian(1.0, THETA, PHI)
        
    else:
        x, y, z = spherical_to_cartesian( np.abs(dr), THETA, PHI )
    #R, THETA, PHI = cartesian_to_spherical(x, y, z)
    
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    #p = ax.plot_surface(x, y, z, cmap=plt.cm.jet, rstride=1, cstride=1, linewidth=0)


    norm = Normalize()
    norm.autoscale(dr)
    colors = cm.jet(norm(dr))
    m = cm.ScalarMappable(cmap=cm.jet, norm=norm)
    m.set_array(dr)
    p = ax.plot_surface(x, y, z, facecolors=colors, rstride=1, cstride=1, linewidth=0)
    plt.colorbar(m, ax=ax)
    
    ax.set_xlabel('$x$')
    ax.set_ylabel('$y$')
    ax.set_zlabel('$z$')
    return fig



"""
Standards
=========

The standards module provides implementations of standards.

.. toctree::
    :maxdepth: 2

    
.. automodule:: acoustics.standards.iso_tr_25417_2007

.. automodule:: acoustics.standards.iso_9613_1_1993

.. automodule:: acoustics.standards.iso_1996_1_2003

.. automodule:: acoustics.standards.iso_1996_2_2007

.. automodule:: acoustics.standards.iec_61672_1_2013

.. automodule:: acoustics.standards.iec_61260_1_2014

"""
from acoustics.standards import (iso_tr_25417_2007, 
                                 iec_61672_1_2013,
                                 iec_61260_1_2014,
                                 iso_9613_1_1993,
                                 iso_1996_1_2003,
                                 iso_1996_2_2007,
                                 )



from __future__ import unicode_literals

import csv
import io
import re

import pandas as pd


def read_csv_cirrus(filename):
    """Read a Cirrus CSV file. Currently exists support for some types of
    CSV files extracted with NoiseTools. There is no support for CSVs related
    with occupational noise.

    If there are NC and NR values in the csv file, they will be stored in the
    returned object with attributes ``nc`` and ``nr``. If the CSV file contains
    time history, you can access to date and time with the ``time`` attribute.
    Also, it is possible to know the integration time with the
    ``integration_time`` attribute.

    :param filename: CSV file name.
    :returns: Pandas dataframe with all data extracted from the CSV file.
    :rtype: Pandas dataframe.

    """
    with open(filename, "r") as csvfile:
        csvreader = csvfile.read()
        csvreader = re.sub(r" dB", "", csvreader)  # Clean " dB" from data
        dialect = csv.Sniffer().sniff(csvreader, delimiters=",;")
        separator = dialect.delimiter
        # Guess decimal separator
        decimal_sep = re.search(r"\"\d{2,3}"
                                r"(\.|,)"  # Decimal separator
                                r"\d{1,2}\"",
                                csvreader).group(1)
    n_cols = re.search("(.+)\n", csvreader).group(1).count(separator) + 1
    if n_cols < 5:
        unsorted_data = []
        pdindex = ["Z"]
        for i, c in enumerate(csvreader.splitlines()):
            if c[:4] == '"NR"':
                nr = int(re.search(r"\d{2}", c).group(0))
                continue
            elif c[:4] == '"NC"':
                nc = int(re.search(r"\d{2}", c).group(0))
                continue
            if i != 0:
                unsorted_data.append(c.split(separator))
            else:
                if n_cols == 3:
                    pdindex.append(c[-2:-1])
                elif n_cols == 4:
                    pdindex.append("A")
                    pdindex.append("C")

        # Create a sorted temporary csv-like file
        csv_data = list(zip(*unsorted_data))
        temp_csv = ""
        for row in csv_data:
            temp_csv += separator.join(row) + "\n"
        # Then, read it with pandas
        data = pd.read_csv(io.StringIO(temp_csv), sep=separator,
                           decimal=decimal_sep)

        # Assign NC and NR data if they are present
        try:
            data.nc = nc
            data.nr = nr
        except:
            pass

        # If the csv file contains global data from the "Details" tab in
        # NoiseTools, skip row names
        if n_cols != 2:
            data.index = pdindex

    else:
        data = pd.read_csv(filename, parse_dates=[[0, 1]], sep=separator,
                           decimal=decimal_sep)

        # Fix time name column
        en_columns = data.columns.values
        en_columns[0] = "time"
        data.columns = en_columns

        # Guess integration time with statistical mode because the csv could
        # have been cleaned from unwanted noise
        data["time"] = pd.to_datetime(data.time)
        delta = data.time.diff().fillna(0)
        # Mode and change from ns to s
        int_time = int(delta.mode().astype(int) * 1e-9)
        if round(int_time, 2) == 0.06:  # Fix for 1/16 s
            int_time = 0.0625
        data.integration_time = int_time

    return data
